/* Written by Jerry James, using code from README.md in the sassy distribution.
 *
 * This code carries the same license as sassy itself; i.e., the MIT license.
 *
 * Copyright (c) 2022 Markus Anders
 * Copyright (c) 2024 Jerry James
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <nauty/naututil.h>
#include <sassy/preprocessor.h>
#include <sassy/tools/nauty_converter.h>

#ifndef _Thread_local
#define _Thread_local thread_local
#endif

static void
hook_function(int n, const int *p, int nsupp, const int *supp)
{
  static int symmetry = 0;
  std::cout << "Symmetry " << ++symmetry << '\n';
  for (int j = 0; j < nsupp; ++j) {
    const int i = supp[j];
    std::cout << i << " -> " << p[i] << '\n';
  }
}

int
main(int argc, char *argv[])
{
  // Create the sassy graph
  sassy::static_graph g;
  g.initialize_graph(8, 12); // 8 vertices, 12 edges
  const int v0 = g.add_vertex(0, 2);
  const int v1 = g.add_vertex(0, 2);
  const int v2 = g.add_vertex(0, 4);
  const int v3 = g.add_vertex(0, 4);
  const int v4 = g.add_vertex(0, 4);
  const int v5 = g.add_vertex(0, 4);
  const int v6 = g.add_vertex(0, 2);
  const int v7 = g.add_vertex(0, 2);
  g.add_edge(v0, v1);
  g.add_edge(v0, v2);
  g.add_edge(v1, v3);
  g.add_edge(v2, v3);
  g.add_edge(v2, v4);
  g.add_edge(v2, v5);
  g.add_edge(v3, v4);
  g.add_edge(v3, v5);
  g.add_edge(v4, v5);
  g.add_edge(v4, v6);
  g.add_edge(v5, v7);
  g.add_edge(v6, v7);

  // Prepare the hook function
  auto hook = sassy::sassy_hook(hook_function);

  // Preprocess the graph
  sassy::preprocessor p;
  p.reduce(&g, &hook);

  // Convert the graph to nauty format
  sparsegraph nauty_graph;
  DYNALLSTAT(int, lab, lab_sz);
  DYNALLSTAT(int, ptn, ptn_sz);
  convert_sassy_to_nauty(&g, &nauty_graph, &lab, &lab_sz, &ptn, &ptn_sz);

  // Pass the graph to nauty
  statsblk stats;
  DYNALLSTAT(int, orbits, orbits_sz);
  DYNALLOC1(int,  orbits, orbits_sz, nauty_graph.nv, "malloc");
  static DEFAULTOPTIONS_SPARSEGRAPH(options);
  options.schreier = true;
  options.defaultptn = false;
  options.userautomproc = sassy::preprocessor::nauty_hook;
  if(nauty_graph.nv > 0) {
    sparsenauty(&nauty_graph, lab, ptn, orbits, &options, &stats, NULL);
  }

  // Clean up
  DYNFREE(lab, lab_sz);
  DYNFREE(ptn, ptn_sz);
  SG_FREE(nauty_graph);

  return 0;
}
